"""Utils for links."""

try:
    from urllib import parse as urlparse
except ImportError:  # pragma: no cover
    import urlparse
import urllib

from flask import current_app


def resolve_url_route(url):
    """Resolve a Flask URL route.

    :param url: `URL`.

    :return: `tuple` of a Flask endpoint and parsed URL parameters.

    :note: This function should be called in Flask application or request context.
    """
    # The real server name is not needed
    url_adapter = current_app.url_map.bind(server_name='')
    return url_adapter.match(urlparse.urlsplit(url).path)


def parse_url_parameters(uri):
    """Parse URI and get the URL parameters.

    :param uri: The resource `URI`.

    :return: A dictionary of the `URL` parameters contained in the `URI`.
    """
    name, parameters = resolve_url_route(uri)
    return parameters


def make_header_link(uri, rel):
    """Format a link according to the HTTP Link specification.

    :param uri: URI of the link.
    :param rel: Relation type.

    :return: A `URI` representing the `Link` header.
    """
    return u'{uri}; rel="{rel}"'.format(uri=uri, rel=rel)


def add_query_params(uri, new_params):
    """Append additional query params to a `URI`.

    :param uri: The `URI` to add query parameters to.
    :param new_params: A dictionary with the parameters to add or update.

    :return: A new `URI` containing query params.
    """
    (scheme, netloc, path, params, query, fragment) = urlparse.urlparse(uri)
    query_params = dict(urlparse.parse_qsl(query, keep_blank_values=True))
    query_params.update(new_params)

    return urlparse.urlunparse(
        (
            scheme,
            netloc,
            path,
            params,
            urllib.urlencode(query_params),
            fragment,
        )
    )



"""Utility functions to map Exceptions to HTTP responses."""

try:
    from http import client as httplib
except ImportError:  # pragma: no cover
    import httplib
import werkzeug.exceptions
from atilla.response import vnd_error_response


def exception_handler(exception):
    """Handle the exceptions and prepares the vnd.error response.

    :param exception: Exception raised in the request/response cycle.
    :return: vnd.error response with the error details.
    """
    if isinstance(exception, werkzeug.exceptions.HTTPException):
        status_code = exception.code or httplib.INTERNAL_SERVER_ERROR
    else:
        status_code = httplib.INTERNAL_SERVER_ERROR

    return vnd_error_response(exception, status_code=status_code)



"""Exceptions for APIs."""

import werkzeug.exceptions


class ApiException(Exception):

    """An exception raised due to user error.

    Exceptions derived from ApiException will be logged automatically if raised.
    User will receive appropriate error response according to the content type.
    """

    description = 'API error.'
    error_type = 'API_ERROR'
    data = None

    def __init__(self, message, description=None, status_code=None, **kwargs):
        """Constructor.

        :param message: the message returned to the user.
        :param description: the message sent to the log.
        """
        self.message = message
        if description is not None or not hasattr(self, 'description'):
            self.description = description

        if status_code is not None:
            self.code = status_code
        if kwargs:
            self.data = kwargs


class NotFound(ApiException, werkzeug.exceptions.NotFound):

    """Not found."""

    description = 'Not found.'
    error_type = 'NOT_FOUND'


class BadRequest(ApiException, werkzeug.exceptions.BadRequest):

    """Bad request."""

    description = 'Bad request.'
    error_type = 'BAD_REQUEST'


class Forbidden(ApiException, werkzeug.exceptions.Forbidden):

    """Forbidden."""

    description = 'Forbidden.'
    error_type = 'FORBIDDEN'


class Unauthorized(ApiException, werkzeug.exceptions.Unauthorized):

    """Unauthorized."""

    description = 'Unauthorized.'
    error_type = 'UNAUTHORIZED'



"""Tests for halogen types."""
import pytest

import halogen

from atilla.ext.halogen import types


class Ticket(halogen.Schema):

    """Test schema."""

    item = halogen.Attr(
        types.URI('static'),
    )


def test_uri(current_app):
    """Test URI schema type."""
    with pytest.raises(halogen.exceptions.ValidationError) as err:
        Ticket.deserialize({'item': ''})
    assert err.value.errors[0].errors[0].args[0] == "'' is not a valid URI for the 'static' endpoint: 404: Not Found"

    with pytest.raises(halogen.exceptions.ValidationError) as err:
        Ticket.deserialize({'item': {}})
    assert err.value.errors[0].errors[0].args[0] == "'{}' is not a valid URI"



"""Flask request and response wrappers."""

try:
    from http import client as httplib
except ImportError:  # pragma: no cover
    import httplib
import json
import halogen

from flask import Request as RequestBase, Response as ResponseBase, abort
from werkzeug.utils import cached_property
from atilla import exceptions


def parse_data(data, mimetype, charset=None):
    """Parse the request/response data.

    :param data: Serialized data.
    :param mimetype: MIME type of the data (only json is supported).
    :param charset: Character set used for the serialization. Default if None.
    :raises: `ValueError` when the data can't be parsed.

    :return: Python objects.
    """
    if data == '' or data is None:
        return

    JSON_MIMETYPES = (
        'application/json',
        'application/hal+json',
    )

    if mimetype in JSON_MIMETYPES:
        if charset is not None:
            return json.loads(data, encoding=charset)
        else:
            return json.loads(data)
    else:
        raise ValueError('Mimetype is not supported')


class Request(RequestBase):

    """Request object used by default in the API.

    The request object is a :class:`flask.wrappers.Request` subclass
    and provides all of the attributes Flask and Werkzeug define plus
    a few API specific ones.
    """

    @cached_property
    def api_key(self):
        """Get the api key."""
        return self.authorization.get('username') if self.authorization else None

    @cached_property
    def parsed_data(self):
        """Parsed the payload data.

        :return: Python object (or list) representing the request payload.

        :raises: Aborts the request with the BAD_REQUEST error.
        """
        try:
            return parse_data(
                data=self.get_data(),
                mimetype=self.mimetype,
                charset=self.mimetype_params.get('charset'),
            )
        except (ValueError, LookupError):
            abort(httplib.BAD_REQUEST, 'The payload could not be parsed')

    @cached_property
    def parsed_form(self):
        """Request form with the list values."""
        if self.method in ('POST', 'PUT'):
            params = self.parsed_data
        else:
            params = {}
            for key, value in self.args.items(multi=True):
                if key in params:
                    param = params.get(key)
                    if isinstance(param, list):
                        param.append(value)
                    else:
                        params[key] = [param, value]
                else:
                    params[key] = value
        return params

    def deserialize(self, schema_class):
        """Deserialize the request data with the schema and validate.

        :param schema_class: Schema class to use for the deserialization and validation.
        :raise: `ValidationError` in case the data doesn't pass the schema validation.
        :return: Deserialized data.
        """
        try:
            return schema_class.deserialize(self.parsed_form or {})
        except halogen.exceptions.ValidationError as e:
            raise exceptions.BadRequest(message='Bad request.', errors=e.to_dict())


class Response(ResponseBase):

    """Response object used by default in the API tests.

    The response object is a :class:`flask.wrappers.Response` subclass
    and provides all of the attributes Flask and Werkzeug define plus
    a few API specific ones.
    """

    @cached_property
    def parsed_data(self):
        """Parsed the response data.

        :return: Python object (or list) representing the request payload.

        :raises: Aborts the request with the BAD_REQUEST error.
        """
        return parse_data(
            data=self.get_data().decode(self.mimetype_params.get('charset', 'utf-8')),
            mimetype=self.mimetype,
            charset=self.mimetype_params.get('charset'),
        )



"""Test exception handlers."""

try:
    from http import client as httplib
except ImportError:  # pragma: no cover
    import httplib
import halogen
import flask


class MySchema(halogen.Schema):

    """Test schema."""

    num = halogen.Attr(halogen.types.Int(validators=[halogen.validators.Range(min=10, max=11)]))


def test_deserialize(application, http):
    """Test validation errors are rendered."""
    @application.route('/')
    def index():
        flask.request.deserialize(MySchema)
        return 'OK'

    response = http.get(
        '/',
        query_string=dict(num=1),
    )
    assert response.status_code == httplib.BAD_REQUEST
    data = response.parsed_data

    assert data['message'] == 'Bad request.'
    assert 'details' in data
    assert 'logref' in data
    assert data['type'] == 'BAD_REQUEST'
    assert application.db._rolledback


def test_unhandled(application, http):
    """Test unhandled exception is being handled."""
    @application.route('/')
    def index():
        1 / 0

    response = http.get('/')
    assert response.status_code == httplib.INTERNAL_SERVER_ERROR
    data = response.parsed_data
    assert 'by zero' in data['message']
    assert 'logref' in data
    assert data['type'] == 'UNKNOWN_ERROR'
    assert application.db._rolledback


def test_unhandled_object(application, http):
    """Test unhandled exception with object as message."""
    @application.route('/')
    def index():

        class SomeObject(object):

            """Some object."""

            def __init__(self, attr):
                self.attr = attr

        raise Exception(SomeObject(1))

    response = http.get('/')
    assert response.status_code == httplib.INTERNAL_SERVER_ERROR
    data = response.parsed_data
    assert data['message'].startswith('Exception: <tests.test_exceptions.')
    assert 'logref' in data
    assert data['type'] == 'UNKNOWN_ERROR'
    assert application.db._rolledback



"""Test configuration."""
import mock
import pytest
from flask.ctx import RequestContext
from flask.wrappers import Request

from atilla.server import create_app


class TestDBSessionManager(object):

    """Test DB session manager."""

    _commited = False
    _rolledback = False
    _closed = False

    do_commit = False

    def commit(self):
        """Mark as committed."""
        self._commited = True

    def close(self):
        """Mark as closed."""
        self._closed = True

    def rollback(self):
        """Mark as rolled back."""
        self._rolledback = True


@pytest.fixture()
def application(mocked_memcached):
    """Atilla Flask test application."""
    app = create_app(
        name='tests',
        mode='test',
        settings_module='tests.settings',
    )
    app.db = TestDBSessionManager()
    return app


@pytest.yield_fixture
def current_app(request, application):
    """Mocking flask.current_app."""
    with application.app_context():
        yield application


@pytest.fixture
def mimetype():
    """Response mimetype."""
    return 'application/hal+json'


@pytest.fixture
def mock_request_context(request, application, mimetype):
    """Mocking flask.current_app."""
    from flask.testing import make_test_environ_builder

    builder = make_test_environ_builder(application)
    environ = builder.get_environ()

    flask_request = Request(environ)
    flask_request.accept_mimetypes = [mimetype]
    request_context = RequestContext(application, environ, request=flask_request)
    request_context.__enter__()

    request.addfinalizer(lambda: request_context.__exit__(None, None, None))


@pytest.yield_fixture
def mocked_memcached():
    """Mocked memcached."""
    mocked = mock.patch('memcache.Client')
    yield mocked.start()
    mocked.stop()


@pytest.fixture
def http(application):
    """Test client.

    Usage:
        with http as c:
            response = c.get(uri)
            print response.parsed_data
    """
    return application.test_client()



"""Caching utils."""

from flask import current_app


def increase_memcached_counter(status_code):
    """Increase the memcached counter for every Response STATUS CODE.

    If the counter does not exist yet, it creates a new one and stores it
    for a number of seconds specified in :ref: `settings.Config.MEMCACHED_COUNTER_TIME`.

    :param status_code: property of flask.Response object.
    :param memcached_client: client to memcached.
    :param namespace: namespace for keys.
    """
    namespace = current_app.config['SERVICE_NAME']

    key = '{namespace}-{status_code}'.format(
        namespace=namespace,
        status_code=status_code
    )
    # .add() only creates the key if it does not exist
    current_app.cache.add(key=key, value='0', timeout=current_app.config['MEMCACHED_COUNTER_TIME'])
    current_app.cache.cache.inc(key=key)



"""Server tests."""
import copy
import logging

from atilla.server import (
    Application,
    create_app,
    httplib,
)


def test_create_app():
    """Test create app shortcut."""
    app = create_app(name='test', mode='test', settings_module='tests.settings')
    assert isinstance(app, Application)


class TestConfig(object):

    """Test application config."""

    LOG_ADDRESS = ('localhost', logging.handlers.SYSLOG_UDP_PORT)


def test_create_app_config():
    """Test create app config."""
    app = create_app(name='test', mode='test', settings_module='tests.test_server')
    config = copy.copy(app.config)
    config.update({
        'BLUEPRINTS': [],
        'SENTRY_DSN': None,
        'MEMCACHED_SERVERS': [],
        'TRANSACTION_COMMIT_STATUS_CODES': [httplib.OK, httplib.CREATED],
        'LOG_FACILITY': 'test',
        'LOG_USER_ERRORS': False,
        'LOG_USER_ERROR_EXC_INFO': False,
        'LOG_USER_ERROR_LEVEL': logging.DEBUG,
        'SENTRY_CAPTURE_USER_ERRORS': False,
        'SERVICE_NAME': 'atilla',
        'MEMCACHED_COUNTER_TIME': 3600,
        'HOST': 'localhost',
        'PORT': '8080',
        'DEBUG': False,
        'LANGUAGES': ['en'],
    })
    assert app.config == config



"""Class/functions for validation."""

import uuid
try:
    from http import client as httplib
except ImportError:  # pragma: no cover
    import httplib

from flask import abort
from werkzeug.routing import BaseConverter


class UidValidator(BaseConverter):

    """Validates the `UID` bit of the route."""

    def to_python(self, value):
        """Check if the value is a valid `UID`.

        :param value: The `UID` value.

        :raises: Abort the request.

        :return: Value as it is.
        """
        try:
            uuid.UUID(value)
        except ValueError:
            abort(httplib.BAD_REQUEST, u'Invalid UID {0}'.format(value))

        return value



"""Test client.

Example usage:

    with app.test_client() as c:
        # Add the content type and authorization headers here
        c.headers = {'Accept': 'application/json', 'Content-Type': 'application/json'}
        response = c.get(url)
"""
from six.moves.urllib import parse as urlparse

from flask import url_for
from flask.testing import FlaskClient


class TestClient(FlaskClient):

    """Flask test client."""

    def open(self, url, *args, **kwargs):
        """Override the open to inject the headers."""
        # Bug in the werkzeug? When the scheme and netloc is present the querystring is ignored
        result = urlparse.urlsplit(url)
        url = urlparse.urlunsplit(('', '', result.path, result.query, result.fragment))

        headers = getattr(self, 'headers', None)
        if headers is not None:
            kwargs.setdefault('headers', headers)
        return super(TestClient, self).open(url, *args, **kwargs)

    def make_uri(self, *args, **kwargs):
        """Reverse the URL in the application context.

        :note: Requires SERVER_NAME to be defined in the configuration.

        """
        with self.application.app_context():
            return url_for(*args, _external=True, **kwargs)



"""Helperfunctions for pagination."""
import collections
import math

from six.moves import http_client
from six.moves.urllib import parse as urlparse
from flask import url_for, current_app

from atilla import exceptions

PageResponse = collections.namedtuple('PageResponse', 'content,count')


def parse_current_page(page):
    """Get page number from page number string.

    :param page: page number
    :type page: str

    :return: `int` parsed page number
    :raises: atilla.exceptions.ApiException - when given page is not an integer or less than 1
    """
    try:
        page = int(page)
    except TypeError:
        message = description = 'Page is not an integer.'
        raise exceptions.ApiException(
            message=message,
            description=description,
            status_code=http_client.BAD_REQUEST,
        )

    if page <= 0:
        message = description = 'The supplied page number is less than 1.'
        raise exceptions.ApiException(
            message=message,
            status_code=http_client.BAD_REQUEST,
            description=description,
        )

    return page


class Page(object):

    """Current page."""

    def __init__(self, func, page):
        """Intialize current page.

        :param func: Function for getting items.
        :param page: Requested page.
        """
        limit_per_page = current_app.config['OBJECTS_PER_PAGE']

        def get_response():
            return func(
                limit=limit_per_page,
                offset=0 if page == 1 else limit_per_page * (page - 1),
            )

        response = get_response()
        self.number_pages = int(math.ceil(float(response.count) / limit_per_page))

        if page > self.number_pages >= 1:
            page = self.number_pages
            response = get_response()

        self.number = page
        self.content = response.content
        self.item_count = response.count

    def has_next_page(self):
        """Helper function for checking that we have next page depends on number of page and number of pages.

        :return: Boolean.
        """
        return self.number < self.number_pages

    def has_prev_page(self):
        """Helper function for checking that we have next page depends on number of page.

        :return: Boolean.
        """
        return self.number > 1


def calculate_first_next_prev_last_links(page, collection_uri):
    """Helper function which updates response with next and prev links.

    :param page: Instance of Page class.
    :param collection_uri: name of the flask route to use as a collection url for pagination links.
    :return: Dictionary with links.
    """
    links = {}
    if page.has_next_page():
        links['next'] = urlparse.urljoin(url_for(collection_uri, _external=True), '?page={0}'.format(page.number + 1))

    if page.has_prev_page():
        links['prev'] = urlparse.urljoin(url_for(collection_uri, _external=True), '?page={0}'.format(page.number - 1))

    if page.number_pages > 1:
        links['last'] = urlparse.urljoin(url_for(collection_uri, _external=True), '?page={0}'.format(page.number_pages))

    if page.number > 1:
        links['first'] = url_for(collection_uri, _external=True)

    return links



"""Tests for atilla/cache.py."""
import mock

from atilla import cache


@mock.patch('werkzeug.contrib.cache.time')
def test_increase_memcached_counter(mocked_time, mocked_memcached, current_app):
    """Increase the memcached counter for every Response STATUS CODE."""
    mocked_time.return_value = 0
    current_app.config['MEMCACHED_COUNTER_TIME'] = 0
    current_app.config['SERVICE_NAME'] = 'namespace'
    cache.increase_memcached_counter('status_code')

    key = 'flask_cache_{namespace}-{status_code}'.format(
        namespace='namespace',
        status_code='status_code'
    )
    mocked_memcached.return_value.add.assert_called_with(key, '0', 0)
    mocked_memcached.return_value.incr.assert_called_with(key, 1)



"""Tests for functions in atilla/response.py."""
import pytest
import werkzeug.exceptions

from atilla import response


def test_get_preferred_accepted_mimetype_with_hal_json(mock_request_context, mimetype):
    """Test that application/hal+json is prefered mimetype"""
    assert response.get_preferred_accepted_mimetype() == mimetype


@pytest.mark.parametrize("mimetype", ["application/json"])
def test_get_preferred_accepted_mimetype_with_json(mock_request_context):
    """Test that application/json is not prefered mimetype"""
    with pytest.raises(werkzeug.exceptions.NotAcceptable):
        response.get_preferred_accepted_mimetype()



"""Tests for functions in atilla/links.py."""
import pytest
from werkzeug import exceptions
from atilla import links


def test_resolve_url_route(current_app):
    """Test resolving a Flask URL route."""

    assert links.resolve_url_route('http://127.0.0.1:8000/static/favicon.ico') == (
        'static', {'filename': 'favicon.ico'}
    )

    with pytest.raises(exceptions.NotFound):
        assert links.resolve_url_route('http://127.0.0.1:8000/shouldnotbefound')


def test_parse_url_parameters(current_app):
    """Test parse URI with the UIDs."""
    uid = 'favicon.ico'
    uri = 'http://127.0.0.1:8000/static/{0}'.format(uid)

    assert links.parse_url_parameters(uri) == {'filename': uid}

    with pytest.raises(exceptions.NotFound):
        assert links.parse_url_parameters('http://127.0.0.1:8000/shouldnotbefound')


def test_make_header_link():
    """Format a link according to the HTTP Link specification."""
    uri = 'http://127.0.0.1:8000/shouldnotbefound'
    rel = 'norel'
    assert links.make_header_link(uri, rel) == u'{uri}; rel="{rel}"'.format(uri=uri, rel=rel)



"""All response related utility functions."""

import logging
try:
    from http import client as httplib
except ImportError:  # pragma: no cover
    import httplib
import json
import uuid

import six

import halogen
from flask import request, abort, current_app, make_response
try:
    from collections import OrderedDict
except ImportError:  # pragma: no cover
    from ordereddict import OrderedDict


MIMETYPE_SERIALIZER = OrderedDict([
    (
        'application/hal+json',
        lambda context: json.dumps(context, indent=4, sort_keys=True)
    ),
])


def api_response(content, status_code=None, headers=None):
    """API response handler.

    :note: The contents are encoded according to the request's accepted mimetype.

    :param content: Response body.
    :param status_code: HTTP status code.
    :param headers: Headers.

    :return: A tuple with `contents`, `status_code`, `headers`.
    """
    mimetype = get_preferred_accepted_mimetype()
    all_headers = set_basic_headers(request, mimetype=mimetype)

    if headers:
        all_headers.update(headers)
    serializer = MIMETYPE_SERIALIZER[mimetype]
    body = serializer(content) if content is not None else ''
    return body, status_code, all_headers


def get_preferred_accepted_mimetype():
    """Return the request's accepted mimetype.

    If the request accepts more than one, it returns the one that appears first on MIMETYPE_SERIALIZER.
    """
    for mimetype, serializer in MIMETYPE_SERIALIZER.items():
        if mimetype in request.accept_mimetypes:
            return mimetype

    abort(
        httplib.NOT_ACCEPTABLE,
        u'The only accepted mimetypes are {0}'.format(', '.join(MIMETYPE_SERIALIZER.keys()))
    )


def get_current_profile_link(*args, **kwargs):
    """Get the profile link from the current endpoint."""
    try:
        schema = current_app.view_functions[request.endpoint].schema
        profile = next((attr for attr in schema.__attrs__ if attr.name == 'profile'))
        return profile.serialize(None)['href']
    except (KeyError, AttributeError, StopIteration):
        return None


def get_error_message(exception):
    """Get the exception message.

    :param exception: Exception object.
    :return: Exception message or description.
    :raises AttributeError: when the message is not available.
    """
    message = getattr(exception, 'message', None)
    if not message:
        message = exception.description
    if not isinstance(message, six.string_types):
        message = u'{0}: {1}'.format(exception.__class__.__name__, message)
    return message


class VNDError(halogen.Schema):

    """An exception that can be serialized to the hal+json format."""

    message = halogen.Attr(required=False, attr=get_error_message)
    logref = halogen.Attr(attr=lambda error: uuid.uuid4().hex)
    help = halogen.Link(required=False, attr=get_current_profile_link)

    # Extended fields
    type = halogen.Attr(required=False, attr='error_type', default='UNKNOWN_ERROR')
    """Error type identifier."""

    details = halogen.Attr(required=False, attr='data')
    """Optional additional error details."""


def vnd_error_response(exception, status_code=None, headers=None):
    """API error response handler.

    If there is an uncaught internal server error it is logged and sent to sentry.
    It is not shown to the users and is not passed along, in essence: it disappears hereafter.

    :param exception: Exception raised in the request/response cycle.
    :param status_code: HTTP status code.
    :param headers: Extra response headers.

    :return: vnd.error response with the error details.
    """
    all_headers = set_basic_headers(request, mimetype='application/hal+json')
    if headers:
        all_headers.update(headers)
    if not hasattr(exception, 'message'):
        try:
            exception.message = exception.args[0]
        except IndexError:
            exception.message = ''
    body = VNDError.serialize(exception)

    if status_code == httplib.INTERNAL_SERVER_ERROR or current_app.config['SENTRY_CAPTURE_USER_ERRORS']:
        sentry = getattr(current_app, 'sentry', None)
        if sentry:
            sentry.captureException()

    if status_code == httplib.INTERNAL_SERVER_ERROR:
        log_exception = True
        exc_info = True
        level = logging.ERROR
    else:
        log_exception = current_app.config['LOG_USER_ERRORS']
        exc_info = current_app.config['LOG_USER_ERROR_EXC_INFO']
        level = current_app.config['LOG_USER_ERROR_LEVEL']

    if log_exception:
        current_app.logger.log(level, u"{type} {logref} {body}".format(body=body, **body), exc_info=exc_info)

    response = make_response(json.dumps(body), status_code)
    response.headers = all_headers
    return response


def set_basic_headers(request, mimetype=None):
    """Set the basic headers for API response.

    :param methods: a wildcard or a list of allowed HTTP methods.
    :param mimetype: a string representing the mimetype to use to render text.

    :return: a dictionary with basic headers.
    """
    headers = {
        'Access-Control-Allow-Headers': ['Content-Type', 'Authorization', 'Accept', 'X-Requested-With'],
        'Access-Control-Expose-Headers': ['Link'],
    }
    if mimetype:
        headers['Content-Type'] = mimetype

    url_adapter = current_app.url_map.bind(
        server_name=current_app.config['SERVER_NAME'] or request.environ['REMOTE_ADDR'])
    methods = url_adapter.allowed_methods(request.path)

    headers['Access-Control-Allow-Methods'] = methods
    headers['Allow'] = methods

    # check the referer because of PREFLIGHT OPTIONS or other AJAX requests
    referer = request.headers.get('Referer')
    if referer:
        origin = request.headers.get('Origin')
        headers['Access-Control-Allow-Credentials'] = 'true'
    else:
        origin = '*'

    headers['Access-Control-Allow-Origin'] = origin

    return headers



"""Manager commands utils."""

from flask.ext.script import Manager, Server, Shell


def add_default_commands(app):
    """Create manager and add shell and runserver commands.

    :return: instance of flask.ext.script.Manager class
    """
    manager = Manager(app)
    manager.add_command('shell', Shell())
    manager.add_command(
        'runserver',
        Server(
            host=app.config['HOST'],
            port=app.config['PORT'],
            use_debugger=app.config['DEBUG'],
            threaded=True,
        )
    )

    return manager



"""Database transaction related functionality."""

import threading


class DatabaseManager(threading.local):

    """Database manager and thread-safe settings."""

    def __init__(self, session, *args, **kwargs):
        """Initialize database manager."""
        self.session = session
        super(DatabaseManager, self).__init__(*args, **kwargs)

    def close(self):
        """Close all database sessions."""
        self.session.close()

    def commit(self):
        """Commit and close the session.

        :note: Rollback the session on error.
        :raise: The exception is re-raised.
        """
        try:
            self.session.commit()
        except Exception:
            self.rollback()
            raise

    def rollback(self):
        """Rollback all database sessions."""
        self.session.rollback()


def create_db_manager(app, session):
    """Create db manager and assign it on app.

    :param session: `Session` object
    """
    if not hasattr(app, 'db'):
        app.db = DatabaseManager(session)



"""Create app stuff."""

try:
    from http import client as httplib
except ImportError:  # pragma: no cover
    import httplib
import logging
import logging.handlers
import os
import sys

from flask import Flask, current_app
from raven.contrib.flask import Sentry
from flask.ext.cache import Cache

from atilla import validators, cache, clients, exception_handlers
from atilla.wrappers import Request, Response

from werkzeug.exceptions import default_exceptions
from werkzeug.utils import import_string


class Application(Flask):

    """Flask application."""

    test_client_class = clients.TestClient
    request_class = Request
    response_class = Response


def load_blueprints(app):
    """Load the blueprints from the configuration."""
    for app_location in app.config['BLUEPRINTS']:
        app_blueprint = import_string(app_location)
        app.register_blueprint(app_blueprint)


def close_sessions_at_start():
    """Close all database sessions at the start of every request."""
    try:
        current_app.db.close()
    except Exception:
        pass


def close_sessions_at_end(*args, **kwargs):
    """Close all database sessions at the end of every request."""
    current_app.db.close()


def create_app(name, mode, app_class=None, config=None, settings_module=None):
    """Create and initialize the application.

    :param name: name of the application
    :type name: str
    :param mode: mode (environment) for the application
    :type mode: str
    :param app_class: optional class for the application object
    :type app_class: type
    :config: application config file name
    :type config: str
    :param settings_module: python settings module for the application
    :type settings_module: str

    :return: application object
    :rtype: flask.Flask
    """
    app = app_class(name) if app_class else Application(name)

    app.config.setdefault('BLUEPRINTS', [])
    app.config.setdefault('SENTRY_DSN', None)
    app.config.setdefault('MEMCACHED_SERVERS', [])
    app.config.setdefault('TRANSACTION_COMMIT_STATUS_CODES', [httplib.OK, httplib.CREATED])
    app.config.setdefault('LOG_FACILITY', 'test')
    app.config.setdefault('LOG_USER_ERRORS', False)
    app.config.setdefault('LOG_USER_ERROR_EXC_INFO', False)
    app.config.setdefault('LOG_USER_ERROR_LEVEL', logging.DEBUG)
    app.config.setdefault('SENTRY_CAPTURE_USER_ERRORS', False)
    app.config.setdefault('SERVICE_NAME', 'atilla')
    app.config.setdefault('MEMCACHED_COUNTER_TIME', 3600)
    app.config.setdefault('HOST', 'localhost')
    app.config.setdefault('PORT', '8080')
    app.config.setdefault('DEBUG', False)
    app.config.setdefault('LANGUAGES', ['en'])
    app.config.setdefault('LOG_ADDRESS', '/var/run/syslog' if sys.platform == 'darwin' else '/dev/log')

    # URL converters
    app.url_map.converters['uid'] = validators.UidValidator

    # Load the configuration
    config_name = u'{0}Config'.format(mode.lower().title())
    try:
        app.config.from_object(
            u'{0}.{1}'.format(settings_module, config_name)
        )
        if config and os.path.exists(config):
            app.config.from_pyfile(config)
    except ImportError:
            raise RuntimeError(u"Can't find {0} in {1}".format(config_name, settings_module))

    # Setup the error handlers
    for code in default_exceptions:
        app.error_handler_spec[None][code] = exception_handlers.exception_handler

    app.register_error_handler(Exception, exception_handlers.exception_handler)

    # Setup Logging
    if app.config['SENTRY_DSN']:
        app.sentry = Sentry(app)
    # Setup cache
    if app.config['MEMCACHED_SERVERS']:
        app.cache = Cache(app, config={
            'CACHE_TYPE': 'memcached',
            'CACHE_MEMCACHED_SERVERS': app.config['MEMCACHED_SERVERS']
        })

    syslog = logging.handlers.SysLogHandler(
        facility=app.config['LOG_FACILITY'],
        address=app.config['LOG_ADDRESS'],
    )
    syslog.setLevel(logging.DEBUG)

    app.logger.addHandler(syslog)

    # Setup endpoints (blueprints)
    with app.app_context():
        load_blueprints(app)

    app.before_request(close_sessions_at_start)
    app.teardown_request(close_sessions_at_end)
    app.after_request(after_request)

    return app


def after_request(response, *args, **kwargs):
    """Log requests and response.

    :param response: flask.Response object.
    :param *args: Required by Flask, but unused.
    :param **kwargs: Required by Flask, but unused.

    :return: flask.Response object.
    """
    if current_app.config['MEMCACHED_SERVERS']:
        cache.increase_memcached_counter(status_code=response.status_code)

    if response.status_code in current_app.config['TRANSACTION_COMMIT_STATUS_CODES']:
        current_app.db.commit()
    else:
        current_app.db.rollback()

    return response



"""Test settings."""
import logging


class TestConfig(object):

    """Test application config."""

    LOG_FACILITY = 'syslog'
    BLUEPRINTS = []
    MEMCACHED_SERVERS = ["127.0.0.1:11211"]
    SERVICE_NAME = 'atilla'
    DEBUG = True
    PORT = 1443
    HOST = '127.0.0.1'
    SERVER_NAME = '{0}:{1}'.format(HOST, PORT)
    SENTRY_DSN = ''
    SENTRY_CAPTURE_USER_ERRORS = True
    LOG_USER_ERRORS = True
    LOG_USER_ERROR_EXC_INFO = True
    LOG_USER_ERROR_LEVEL = logging.INFO
    MEMCACHED_COUNTER_TIME = 0
    TRANSACTION_COMMIT_STATUS_CODES = [200, 201]
    OBJECTS_PER_PAGE = 2
    LOG_ADDRESS = ('localhost', logging.handlers.SYSLOG_UDP_PORT)



"""Tests for pagination helpers."""
import pytest

from atilla.pagination import Page, PageResponse


@pytest.mark.parametrize('items, page, has_next_page, has_prev_page', [
    ([1, 2, 3], 1, True, False),
    ([1, 2, 3], 2, False, True),
    ([1, 2, 3, 4, 5], 2, True, True),
])
def test_page(current_app, items, page, has_next_page, has_prev_page):
    """Test Page helper."""
    page = Page(lambda limit, offset: PageResponse(
        content=items[offset:offset + limit],
        count=len(items)
    ), page)
    assert page.has_next_page() is has_next_page
    assert page.has_prev_page() is has_prev_page



"""Tests for validators."""

import pytest
import uuid
import werkzeug.exceptions

from atilla import validators


def test_uid_validator():
    """Check that UidValidator not fail with normal uuid."""
    random_uuid = str(uuid.uuid4())
    uid_validator = validators.UidValidator({})
    assert uid_validator.to_python(random_uuid) == random_uuid


def test_uid_validator_raises_exception():
    """Check that UidValidator not fail with normal uuid."""
    uid_validator = validators.UidValidator({})
    with pytest.raises(werkzeug.exceptions.BadRequest):
        uid_validator.to_python('wrong_uuid')



"""API serialization."""

import functools
from atilla import response


def serialize(schema_class):
    """Serialize the API call result with the schema.

    :param schema_class: Schema class to use for the serialization.

    :return: Decorator that applies the serialization to a function.
    """
    def wrapper(func):

        func.schema = schema_class

        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            resp = func(*args, **kwargs)
            params = {}
            context = {}
            if type(resp) == tuple:
                data, context = resp[:2]
                if len(resp) == 3:
                    params = resp[2]
            else:
                data = resp
            return response.api_response(schema_class.serialize(data, **context), **params)
        return func_wrapper

    return wrapper



"""Schema types."""
import six

import werkzeug.exceptions
import flask
import halogen
import halogen.exceptions

from atilla import links


class URI(halogen.types.Type):

    """Resource id to URI conversion type."""

    def __init__(self, endpoint, external=True):
        """`URI` type constructor.

        :param endpoint: Flask route endpoint name to compare with when specified.
        :param external: Is URI external (include schema and domain).
        """
        self.endpoint = endpoint
        self.external = external
        super(URI, self).__init__()

    def serialize(self, value):
        """Convert resource id to :ref:`URI`.

        :param value: Resource id value.
        :return: :ref:`URI`.
        """
        if not isinstance(value, dict):
            value = {'uid': value}

        return flask.url_for(self.endpoint, _external=self.external, **value)

    def deserialize(self, value):
        """Parse :ref:`URI` and get resource id.

        :param value: :ref:`URI` value.
        :return: Converted resource id.
        """
        if not isinstance(value, six.string_types):
            raise ValueError(u"'{val}' is not a valid URI".format(val=value))
        try:
            endpoint, params = links.resolve_url_route(value)
            if self.endpoint:
                assert endpoint == self.endpoint
        except (RuntimeError, AssertionError):  # Re-raise the Flask context exception
            raise
        except werkzeug.exceptions.HTTPException as e:
            raise ValueError(
                u"'{val}' is not a valid URI for the '{endpoint}' endpoint: {err}".format(
                    val=value,
                    endpoint=self.endpoint,
                    err=e,
                )
            )
        try:
            return params.values()[0] if len(params) == 1 else params
        except (TypeError, IndexError):
            raise ValueError(u'Could not parse the URI parameters {0}.'.format(unicode(params)))



"""Setuptools entry point."""
import codecs
import os.path

import setuptools

# Get the version number from the code base.
from atilla import __version__ as atilla_version

tests_require = [
    "isodate",
    "mock",
    "pytest",
    "pytest-cov",
    "pytest-pep8",
    "pytest-pep257",
    "pytz",
]

long_description = []

for text_file in ['README.rst', 'CHANGES.rst']:
    with codecs.open(os.path.join(os.path.dirname(os.path.abspath(__file__)), text_file), encoding='utf-8') as f:
        long_description.append(f.read())

install_requires = [
        "argparse>=1.2.1",
        "flask-script>=0.4.0",
        "flask>=0.10.1",
        "flask-cache",
        "halogen>=1.0.5",
        "python-memcached",
        "raven[flask]",
        "six",
        "Werkzeug>=0.9.3",
    ]

try:
    from collections import OrderedDict  # NOQA
except ImportError:  # pragma: no cover
    install_requires.append('ordereddict')

setuptools.setup(
    name='atilla',
    description="flask API projects helper",
    packages=setuptools.find_packages(include="atilla*"),
    long_description='\n'.join(long_description),
    author="Paylogic International and others",
    license="MIT license",
    author_email="developers@paylogic.com",
    url="https://github.com/paylogic/atilla",
    version=atilla_version,
    classifiers=[
        "Development Status :: 6 - Mature",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3"
    ] + [("Programming Language :: Python :: %s" % x) for x in "2.7 3.4".split()],
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        "test": tests_require,
    }
)



"""Utility functions."""

from flask import current_app


def to_i18n_dict(messages):
    """Convert to I18N dictionary.

    Copies the same value for all the provided languages, and creates empty value for the non provided ones.

    :param messages: A `dict` of messages to convert to i18n object for all available languages.
    :type messages: `dict`

    :return: `dict` containing the message for the all the supported languages.
    """
    return dict((lang, messages.get(lang, u"") if messages else u"") for lang in current_app.config['LANGUAGES'])
